'''
Asyncio equivalents to regular Python functions.

'''
import asyncio
import itertools

from . import types


async def acount(
    start: float = 0,
    step: float = 1,
    delay: float = 0,
    stop: types.Optional[float] = None,
) -> types.AsyncIterator[float]:
    '''Asyncio version of itertools.count()'''
    for item in itertools.count(start, step):  # pragma: no branch
        if stop is not None and item >= stop:
            break

        yield item
        await asyncio.sleep(delay)



from unittest.mock import MagicMock

import pytest

from python_utils.decorators import sample


@pytest.fixture
def random(monkeypatch):
    mock = MagicMock()
    monkeypatch.setattr(
        "python_utils.decorators.random.random", mock, raising=True
    )
    return mock


def test_sample_called(random):
    demo_function = MagicMock()
    decorated = sample(0.5)(demo_function)
    random.return_value = 0.4
    decorated()
    random.return_value = 0.0
    decorated()
    args = [1, 2]
    kwargs = {"1": 1, "2": 2}
    decorated(*args, **kwargs)
    demo_function.assert_called_with(*args, **kwargs)
    assert demo_function.call_count == 3


def test_sample_not_called(random):
    demo_function = MagicMock()
    decorated = sample(0.5)(demo_function)
    random.return_value = 0.5
    decorated()
    random.return_value = 1.0
    decorated()
    assert demo_function.call_count == 0



from __future__ import annotations

import loguru

from . import logger as logger_module

__all__ = ['Logurud']


class Logurud(logger_module.LoggerBase):
    logger: loguru.Logger

    def __new__(cls, *args, **kwargs):
        cls.logger: loguru.Logger = loguru.logger.opt(depth=1)
        return super().__new__(cls)



import asyncio
import itertools
from datetime import timedelta

import pytest

import python_utils


@pytest.mark.parametrize(
    'timeout,interval,interval_multiplier,maximum_interval,iterable,result',
    [
        (0.2, 0.1, 0.4, 0.2, python_utils.acount, 2),
        (0.3, 0.1, 0.4, 0.2, python_utils.acount(), 3),
        (0.3, 0.06, 1.0, None, python_utils.acount, 5),
        (
            timedelta(seconds=0.1),
            timedelta(seconds=0.06),
            2.0,
            timedelta(seconds=0.1),
            python_utils.acount,
            2,
        ),
    ],
)
@pytest.mark.asyncio
async def test_aio_timeout_generator(
    timeout, interval, interval_multiplier, maximum_interval, iterable, result
):
    i = None
    async for i in python_utils.aio_timeout_generator(
        timeout, interval, iterable, maximum_interval=maximum_interval
    ):
        pass

    assert i == result


@pytest.mark.parametrize(
    'timeout,interval,interval_multiplier,maximum_interval,iterable,result',
    [
        (0.01, 0.006, 0.5, 0.01, 'abc', 'c'),
        (0.01, 0.006, 0.5, 0.01, itertools.count, 2),
        (0.01, 0.006, 0.5, 0.01, itertools.count(), 2),
        (0.01, 0.006, 1.0, None, 'abc', 'c'),
        (
            timedelta(seconds=0.01),
            timedelta(seconds=0.006),
            2.0,
            timedelta(seconds=0.01),
            itertools.count,
            2,
        ),
    ],
)
def test_timeout_generator(
    timeout, interval, interval_multiplier, maximum_interval, iterable, result
):
    i = None
    for i in python_utils.timeout_generator(
        timeout=timeout,
        interval=interval,
        interval_multiplier=interval_multiplier,
        iterable=iterable,
        maximum_interval=maximum_interval,
    ):
        pass

    assert i == result


@pytest.mark.asyncio
async def test_aio_generator_timeout_detector():
    # Make pyright happy
    i = None

    async def generator():
        for i in range(10):
            await asyncio.sleep(i / 100.0)
            yield i

    detector = python_utils.aio_generator_timeout_detector
    # Test regular timeout with reraise
    with pytest.raises(asyncio.TimeoutError):
        async for i in detector(generator(), 0.05):
            pass

    # Test regular timeout with clean exit
    async for i in detector(generator(), 0.05, on_timeout=None):
        pass

    assert i == 4

    # Test total timeout with reraise
    with pytest.raises(asyncio.TimeoutError):
        async for i in detector(generator(), total_timeout=0.1):
            pass

    # Test total timeout with clean exit
    async for i in detector(generator(), total_timeout=0.1, on_timeout=None):
        pass

    assert i == 4

    # Test stop iteration
    async for i in detector(generator(), on_timeout=None):
        pass


@pytest.mark.asyncio
async def test_aio_generator_timeout_detector_decorator():
    # Make pyright happy
    i = None

    # Test regular timeout with reraise
    @python_utils.aio_generator_timeout_detector_decorator(timeout=0.05)
    async def generator_timeout():
        for i in range(10):
            await asyncio.sleep(i / 100.0)
            yield i

    with pytest.raises(asyncio.TimeoutError):
        async for i in generator_timeout():
            pass

    # Test regular timeout with clean exit
    @python_utils.aio_generator_timeout_detector_decorator(
        timeout=0.05, on_timeout=None
    )
    async def generator_clean():
        for i in range(10):
            await asyncio.sleep(i / 100.0)
            yield i

    async for i in generator_clean():
        pass

    assert i == 4

    # Test total timeout with reraise
    @python_utils.aio_generator_timeout_detector_decorator(total_timeout=0.1)
    async def generator_reraise():
        for i in range(10):
            await asyncio.sleep(i / 100.0)
            yield i

    with pytest.raises(asyncio.TimeoutError):
        async for i in generator_reraise():
            pass

    # Test total timeout with clean exit
    @python_utils.aio_generator_timeout_detector_decorator(
        total_timeout=0.1, on_timeout=None
    )
    async def generator_clean_total():
        for i in range(10):
            await asyncio.sleep(i / 100.0)
            yield i

    async for i in generator_clean_total():
        pass

    assert i == 4



import pytest

from python_utils import containers


def test_unique_list_ignore():
    a = containers.UniqueList()
    a.append(1)
    a.append(1)
    assert a == [1]

    a = containers.UniqueList(*range(20))
    with pytest.raises(RuntimeError):
        a[10:20:2] = [1, 2, 3, 4, 5]

    a[3] = 5


def test_unique_list_raise():
    a = containers.UniqueList(*range(20), on_duplicate='raise')
    with pytest.raises(ValueError):
        a[10:20:2] = [1, 2, 3, 4, 5]

    a[10:20:2] = [21, 22, 23, 24, 25]
    with pytest.raises(ValueError):
        a[3] = 5

    del a[10]
    del a[5:15]
